# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import re
import lxml.etree
#import traceback

# ----------------------------------------------------------------------

try:
    unicode         # Python 2.x
    def lxml_element2string(element, method="text", with_tail=False):
        return lxml.etree.tostring(element, method=method,
                encoding=unicode, with_tail=with_tail)
except NameError:   # Python 3.x
    def lxml_element2string(element, method="text", with_tail=False):
        return lxml.etree.tostring(element, method=method,
                encoding=str, with_tail=with_tail)
except:
    raise

def extract_text(element, keep_nl=False, with_tail=False):
    return remove_multiple_whitespaces(
        lxml_element2string(element, method="text", with_tail=with_tail),
        keep_nl=keep_nl).strip()

def extract_html(element, with_tail=False):
    return lxml_element2string(element, method="html", with_tail=with_tail)

def extract_xml(element, with_tail=False):
    return lxml_element2string(element, method="xml", with_tail=with_tail)

REGEX_NEWLINE = re.compile(r'\n')
REGEX_WHITESPACE = re.compile(r'\s+', re.UNICODE)
def remove_multiple_whitespaces(input_string, keep_nl=False):

    if keep_nl:
        lines = REGEX_NEWLINE.split(input_string)
        return "\n".join([remove_multiple_whitespaces(l) for l in lines])
    else:
        return REGEX_WHITESPACE.sub(" ", input_string).strip()


def format_alter_htmltags(tree, tags=[], replacement=" "):
    regex_repl_start = re.compile(r'^\s*%s' % replacement, re.UNICODE)
    context = lxml.etree.iterwalk(tree, events=("end", ))
    tag_set = set(tags)
    for action, elem in context:
        if elem.tag not in tag_set:
            continue
        if elem.tail is None:
            elem.tail = replacement
        elif not regex_repl_start.search(elem.tail):
            elem.tail = "%s%s" % (replacement, elem.tail)
    return tree


HTML_BLOCK_ELEMENTS = [
    'address',
    'article',
    'aside',
    'audio',
    'blockquote',
    'br',
    'canvas',
    'div',
    'dl', 'dd', 'dt',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    'header',
    'hgroup',
    'hr',
    'noscript',
    'li', 'ol', 'ul',
    'output',
    'p',
    'pre',
    'section',
    'table',
    'tfoot',
    'video',
]
def format_htmlblock_tags(tree, replacement="\n"):
    return format_alter_htmltags(tree,
        tags=HTML_BLOCK_ELEMENTS,
        replacement=replacement)


def elements2text(nodes, with_tail=True):
    return [extract_text(e, with_tail=with_tail) for e in nodes]


def elements2textnl(nodes, with_tail=True, replacement="\n"):
    return [extract_text(
                format_htmlblock_tags(e, replacement=replacement),
                with_tail=with_tail,
                keep_nl=True)
            for e in nodes]

def elements2html(nodes):
    return [extract_html(e) for e in nodes]

def elements2xml(nodes):
    return [extract_xml(e) for e in nodes]

# ----------------------------------------------------------------------

def test_listitems_type(itemlist, checktype):
    return all([isinstance(i, checktype) for i in itemlist])

def check_listitems_types(itemlist):
    return list(set([type(i) for i in itemlist]))

def apply2elements(elements, element_func, notelement_func=None):
    ltype = check_listitems_types(elements)
    if ltype == [lxml.etree._Element]:
        return element_func(elements)
    elif notelement_func:
        return notelement_func(elements)
    else:
        return elements

#def apply2element(element, element_func, notelement_func=None):
    #if type(element) == lxml.etree._Element:
        #return element_func(element)
    #elif notelement_func:
        #return notelement_func(element)
    #else:
        #return element

try:
    unicode         # Python 2.x
    def xpathtostring(context, nodes, with_tail=True, *args):
        return apply2elements(
            nodes,
            element_func=lambda nodes: elements2text(
                nodes, with_tail=with_tail),
            notelement_func=lambda nodes: [
                remove_multiple_whitespaces(unicode(s))
                    for s in nodes],
        )

except NameError:   # Python 3.x
    def xpathtostring(context, nodes, with_tail=True, *args):
        return apply2elements(
            nodes,
            element_func=lambda nodes: elements2text(
                nodes, with_tail=with_tail),
            notelement_func=lambda nodes: [
                remove_multiple_whitespaces(str(s))
                    for s in nodes],
        )

def xpathtostringnl(context, nodes, with_tail=True, replacement="\n", *args):
    return apply2elements(nodes,
        element_func=lambda nodes: elements2textnl(
            nodes, with_tail=with_tail, replacement=replacement))

def xpathtohtml(context, nodes):
    return apply2elements(nodes,
        element_func=lambda nodes: elements2html(nodes))

def xpathtoxml(context, nodes):
    return apply2elements(nodes,
        element_func=lambda nodes: elements2xml(nodes))

try:
    unicode         # Python 2.x
    def xpathstrip(context, nodes, stripchars=None, with_tail=True, *args):
        if test_listitems_type(nodes, lxml.etree._Element):
            return [s.strip(stripchars)
                    for s in elements2text(
                        nodes, with_tail=with_tail)]
        else:
            return [unicode(s).strip(stripchars) for s in nodes]

except NameError:   # Python 3.x
    def xpathstrip(context, nodes, stripchars=None, with_tail=True, *args):
        if test_listitems_type(nodes, lxml.etree._Element):
            return [s.strip(stripchars)
                    for s in elements2text(
                        nodes, with_tail=with_tail)]
        else:
            return [str(s).strip(stripchars) for s in nodes]


def xpathattrname(context, attributes, *args):
    return [a.attrname for a in attributes]



import global_settings


class Settings(object):
    def __init__(self, settings_module=None):
        for setting in dir(global_settings):
            setting_value = getattr(global_settings, setting)
            setattr(self, setting, setting_value)
        if settings_module:
            self.update(settings_module)

    def update(self, module):
        for setting in dir(module):
            if setting.isupper() and hasattr(self, setting):
                new_value = getattr(module, setting)
                old_value = getattr(self, setting)
                if isinstance(old_value, dict):
                    old_value.update(new_value)
                    new_value = old_value
                setattr(self, setting, new_value)



from gevent import spawn, joinall, monkey
monkey.patch_all()

from dragline import __version__, runtime
import argparse
import os
import logging
from .crawl import Crawler
from .settings import Settings
from logging.config import dictConfig
from importlib import import_module
from utils import load_module


def get_request_processor(processor):
    module, classname = processor.split(':')
    return getattr(import_module(module), classname)()


def configure_runtime(spider, settings):
    runtime.settings = settings
    dictConfig(settings.LOGGING)
    runtime.request_processor = get_request_processor(settings.REQUEST_PROCESSOR)
    runtime.spider = spider
    if hasattr(settings, 'NAMESPACE'):
        runtime.logger = logging.getLogger(str(settings.NAMESPACE))
        runtime.logger = logging.LoggerAdapter(runtime.logger, {"spider_name": spider.name})
    else:
        runtime.logger = logging.getLogger(spider.name)
    spider.logger = runtime.logger


def main(spider, settings):
    if not isinstance(settings, Settings):
        settings = Settings(settings)
    configure_runtime(spider, settings)
    crawler = Crawler()
    threads = runtime.settings.THREADS
    try:
        joinall([spawn(crawler.process_url) for i in xrange(threads)])
    except KeyboardInterrupt:
        crawler.clear(False)
    except:
        runtime.logger.exception("Unable to complete")
    else:
        crawler.clear(True)
        runtime.logger.info("Crawling completed")


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('spider', help='spider directory name')
    parser.add_argument('--resume', '-r', action='store_true',
                        help="resume crawl")
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    args = parser.parse_args()
    path = os.path.abspath(args.spider)
    spider = load_module(path, 'main').Spider
    settings = load_module(path, 'settings')
    if args.resume:
        settings.RESUME = True
    main(spider(), settings)

if __name__ == "__main__":
    run()



from __future__ import absolute_import
from Queue import Queue, Empty
from selenium import webdriver
from datetime import timedelta
from dragline import runtime
from selenium.webdriver.common.proxy import Proxy, ProxyType
from selenium.webdriver.firefox.firefox_profile import FirefoxProfile
class Driver(object):

    def __len__(self):
        return 0

    @property
    def status(self):
        return 200

    @property
    def content(self):
        return self.page_source

    @property
    def url(self):
        return self.current_url

    @property
    def elapsed(self):
        return timedelta()


class Remote(webdriver.Remote, Driver):
    pass


class FirefoxDriver(webdriver.Firefox, Driver):
    pass


class ChromeDriver(webdriver.Chrome, Driver):
    pass


class PhantomJSDriver(webdriver.PhantomJS, Driver):
    pass


class Browser(object):
    def get_driver(self, **kwargs):
        return Remote(**kwargs)

    def __init__(self):
        self.browsers = Queue()

    def get_response(self, url, **kwargs):
        try:
            browser = self.browsers.get(block=False)
        except Empty:
            proxy = runtime.settings.SELENIUM_ARGS.get('proxy')
            if proxy:
                proxy = Proxy({
                    'proxyType': ProxyType.MANUAL,
                    'httpProxy': proxy,
                })
                browser = self.get_driver(proxy=proxy)
            else:
                browser = self.get_driver()
            if hasattr(runtime.spider,"init_browser"):
                runtime.spider.init_browser(browser)
        width = runtime.settings.SELENIUM_ARGS.get('WINDOW_SIZE_WIDTH')
        height = runtime.settings.SELENIUM_ARGS.get('WINDOW_SIZE_HEIGHT')
        browser.set_window_size(width, height)
        browser.get(url)
        return browser

    def put_response(self, browser):
        self.browsers.put(browser)

    def clear(self):
        while True:
            try:
                browser = self.browsers.get(block=False)
                browser.close()
            except Empty:
                break


class Headless(Browser):
    def __init__(self):
        from xvfbwrapper import Xvfb
        width = runtime.settings.SELENIUM_ARGS.get('WINDOW_SIZE_WIDTH')
        height = runtime.settings.SELENIUM_ARGS.get('WINDOW_SIZE_HEIGHT')
        self._vdisplay = Xvfb(width=width, height=height)
        self._vdisplay.start()
        super(Headless, self).__init__()

    def clear(self):
        super(Headless, self).clear()
        self._vdisplay.stop()


class Chrome(Browser):
    def get_driver(self, **kwargs):
        return ChromeDriver(**kwargs)


class ChromeX(Headless, Chrome):
    pass


class Firefox(Browser):
    def get_driver(self, **kwargs):
        firefoxProfile = FirefoxProfile()
        # Disable CSS
        if runtime.settings.SELENIUM_ARGS.get('DISABLE_CSS'):
            firefoxProfile.set_preference('permissions.default.stylesheet', 2)
        # Disable images
        if runtime.settings.SELENIUM_ARGS.get('DISABLE_IMAGE'):
            firefoxProfile.set_preference('permissions.default.image', 2)
        # Disable Flash
        if runtime.settings.SELENIUM_ARGS.get('DISABLE_FLASH'):
            firefoxProfile.set_preference('dom.ipc.plugins.enabled.libflashplayer.so', 'false')
        kwargs['firefox_profile'] = firefoxProfile
        return FirefoxDriver(**kwargs)


class FirefoxX(Headless, Firefox):
    pass


class PhantomJS(Browser):
    def get_driver(self, **kwargs):
        return PhantomJSDriver(**kwargs)



from dragline import __version__, runtime
try:
    from cPickle import Pickler, Unpickler, HIGHEST_PROTOCOL
except:
    from pickle import Pickler, Unpickler, HIGHEST_PROTOCOL
from . import redisds
from gevent.lock import BoundedSemaphore
from .http import Request, RequestError
from uuid import uuid4
from datetime import datetime
from pytz import timezone
import time
from requests.compat import urlsplit
try:
    from cStringIO import StringIO
except ImportError:
    from requests.compat import StringIO


class Pickle():

    def dumps(self, obj, protocol=HIGHEST_PROTOCOL):
        file = StringIO()
        Pickler(file, protocol).dump(obj)
        return file.getvalue()

    def loads(self, str):
        file = StringIO(str)
        return Unpickler(file).load()


class Crawler:
    def __init__(self):
        self.load()

    def load(self):
        redis_args = dict(host=runtime.settings.REDIS_URL,
                          port=runtime.settings.REDIS_PORT,
                          db=runtime.settings.REDIS_DB)
        if hasattr(runtime.settings, 'NAMESPACE'):
            redis_args['namespace'] = runtime.settings.NAMESPACE
        else:
            redis_args['namespace'] = runtime.spider.name
        self.url_set = redisds.Set('urlset', **redis_args)
        self.url_queue = redisds.Queue('urlqueue', serializer=Pickle(),
                                       **redis_args)
        self.runner = redisds.Lock("runner:%s" % uuid4().hex, **redis_args)
        self.runners = redisds.Dict("runner:*", **redis_args)
        self.publiser = redisds.Publiser(**redis_args)
        runtime.stats = redisds.Hash("stats", **redis_args)
        self.conf = redisds.Hash("conf", **redis_args)
        self.lock = BoundedSemaphore(1)
        self.running_count = 0
        if not hasattr(runtime.spider, 'allowed_domains'):
            runtime.spider.allowed_domains = []
        self.start()

    def current_time(self):
        tz = timezone(runtime.settings.TIME_ZONE)
        return datetime.now(tz).isoformat()

    def start(self):
        self.conf['DELAY'] = runtime.settings.MIN_DELAY
        if not runtime.settings.RESUME and self.is_inactive():
            self.url_queue.clear()
            self.url_set.clear()
            runtime.stats.clear()
        if isinstance(runtime.spider.start, list):
            requests = runtime.spider.start
        else:
            requests = [runtime.spider.start]
        for request in requests:
            if isinstance(request, basestring):
                request = Request(request)
            if request.callback is None:
                request.callback = "parse"
            self.insert(request)
        if runtime.stats.setnx('status', "running") or runtime.stats.setifval('status', "stopped", "running"):
            runtime.stats['start_time'] = self.current_time()
            runtime.logger.info("Starting spider %s", dict(iter(runtime.stats)))
            self.publiser.publish('status_changed:running')
        else:
            runtime.logger.info("Supporting %s", dict(iter(runtime.stats)))

    def clear(self, finished):
        self.runner.release()
        status = 'finished' if finished else 'stopped'
        if self.is_inactive() and runtime.stats.setifval('status', 'running', status):
            runtime.stats['end_time'] = self.current_time()
            if finished:
                self.url_queue.clear()
                self.url_set.clear()
            runtime.logger.info("%s", dict(iter(runtime.stats)))
            self.publiser.publish('status_changed:stopped')
        runtime.request_processor.clear()

    def is_inactive(self):
        return len(self.runners) == 0

    def inc_count(self):
        self.lock.acquire()
        if self.running_count == 0:
            self.runner.acquire()
        self.running_count += 1
        self.lock.release()

    def decr_count(self):
        self.lock.acquire()
        self.running_count -= 1
        if self.running_count == 0:
            self.runner.release()
        self.lock.release()

    def insert(self, request, check=True):
        if not isinstance(request, Request):
            return
        url = urlsplit(request.url)
        if not all((url.scheme in ['http', 'https'], url.hostname)):
            runtime.logger.debug('invalid url %s', url.geturl())
            return
        reqhash = request.get_unique_id(True)
        if check:
            check = not request.dont_filter
        if check:
            if runtime.spider.allowed_domains and url.hostname not in runtime.spider.allowed_domains:
                runtime.logger.debug('invalid url %s (domain %s not in %s)', url.geturl(), url.hostname, str(runtime.spider.allowed_domains))
                return
            elif runtime.settings.UNIQUE_CHECK:
                if not self.url_set.add(reqhash):
                    return
        self.url_queue.put(request)
        del request

    def updatedelay(self, delay):
        self.conf['DELAY'] = min(
            max(runtime.settings.MIN_DELAY, delay,
                (float(self.conf['DELAY']) + delay) / 2.0),
            runtime.settings.MAX_DELAY)

    def process_request(self, request):
        response = None
        try:
            response = request.send()
            if runtime.settings.AUTOTHROTTLE:
                self.updatedelay(response.elapsed.seconds)
                time.sleep(float(self.conf['DELAY']))
            runtime.stats.inc('pages_crawled')
            runtime.stats.inc("status_code:" + str(response.status))
            if len(response):
                runtime.stats.inc('request_bytes', len(response))
            requests = request.callback(response)
            if requests:
                for i in requests:
                    self.insert(i)
        except:
            raise
        finally:
            if response is not None:
                runtime.request_processor.put_response(response)

    def process_url(self):
        while runtime.stats['status'] == "running":
            request = self.url_queue.get(timeout=2)
            if request:
                runtime.logger.debug("Processing %s", request)
                self.inc_count()
                try:
                    self.process_request(request)
                except RequestError:
                    request.retry += 1
                    runtime.stats.inc('retry_count')
                    if request.retry >= runtime.settings.MAX_RETRY:
                        runtime.logger.warning("Rejecting %s and meta is %s", request, str(request.meta), exc_info=True)
                    else:
                        runtime.logger.debug("Retrying %s", request, exc_info=True)
                        self.insert(request, False)
                except KeyboardInterrupt:
                    self.insert(request, False)
                    raise KeyboardInterrupt
                except:
                    runtime.logger.exception("Failed to execute callback on %s and meta is %s", request, str(request.meta))
                else:
                    runtime.logger.info("Finished processing %s", request)
                finally:
                    self.decr_count()
            else:
                if self.is_inactive():
                    break
                runtime.logger.debug("No url to process, active threads: %s", self.running_count)



import redis
import time
import uuid
import threading


class RedisPoolManager:

    def __init__(self):
        self.pools = {}

    def getpool(self, host='localhost', port=6379, db=0):
        url = "redis://%s:%s/%d" % (host, port, db)
        if url not in self.pools:
            self.pools[url] = redis.BlockingConnectionPool.from_url(url)
        return self.pools[url]

poolmanager = RedisPoolManager()


class Publiser(object):

    """Simple Queue with Redis Backend"""

    def __init__(self, namespace='signal', **redis_kwargs):
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.key = namespace

    def publish(self, channel):
        self.__db.publish(channel, self.key)


class Queue(object):

    """Simple Queue with Redis Backend"""

    def __init__(self, name, namespace='queue', serializer=None,
                 **redis_kwargs):
        """
        The default parameters are:
            namespace='queue', serializer=None, hash_func=usha1
            host='localhost', port=6379, db=0
        """
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.key = '%s:%s' % (namespace, name)
        self.serializer = serializer

    def __len__(self):
        return self.qsize()

    def clear(self):
        """Deletes a list"""
        self.__db.delete(self.key)

    def qsize(self):
        """Return the approximate size of the queue."""
        return self.__db.llen(self.key)

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return self.qsize() == 0

    def remove(self, item):
        """Remove all elements in the queue equal to item."""
        if self.serializer:
            item = self.serializer.dumps(item)
        return self.__db.lrem(self.key, item, 0)

    def put(self, item):
        """Put item into the queue."""
        if self.serializer:
            item = self.serializer.dumps(item)
        self.__db.rpush(self.key, item)

    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args block is true and timeout is None (the default), block
        if necessary until an item is available."""
        if block:
            item = self.__db.blpop(self.key, timeout=timeout)
            if item:
                item = item[1]
        else:
            item = self.__db.lpop(self.key)
        if item:
            if self.serializer:
                item = self.serializer.loads(item)
        return item

    def get_nowait(self):
        """Equivalent to get(False)."""
        return self.get(False)


class Dict(object):

    """Simple Dict with Redis Backend"""

    def __init__(self, pattern='*', namespace='dict', **redis_kwargs):
        """
        The default parameters are:
            namespace='dict', host='localhost', port=6379, db=0
        """
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.pattern = '%s:%s' % (namespace, pattern)

    def __len__(self):
        return len(list(self.__db.scan_iter(self.pattern)))

    def __setitem__(self, idx, value):
        name = self.pattern.replace('*', idx)
        self.__db.set(name, value)

    def __getitem__(self, key):
        name = self.pattern.replace('*', key)
        if self.__db.type(name) not in ['string', 'none']:
            return None
        value = self.__db.get(name)
        if value is None:
            return 0
        if value.isdigit():
            return int(value)
        return value

    def inc(self, key, value=1):
        name = self.pattern.replace('*', key)
        return self.__db.incr(name, value)

    def __iter__(self):
        for n, i in enumerate(self.pattern.split(':')):
            if i == "*":
                break
        keys = [key.split(':')[n] for key in self.__db.scan_iter(self.pattern)]
        return ((key, self[key]) for key in keys)

    def clear(self):
        val = list(self.__db.scan_iter(self.pattern))
        if val:
            return self.__db.delete(*val)


class Set(object):

    "A simple set with redis backend"

    def __init__(self, name, namespace='set', **redis_kwargs):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.key = '%s:%s' % (namespace, name)

    def __len__(self):
        "returns the size of set"
        return self.__db.scard(self.key)

    def empty(self):
        "check whether the set is empty"
        return len(self) == 0

    def add(self, item):
        "Add an item to the set"
        return self.__db.sadd(self.key, item) == 1

    def __contains__(self, item):
        return self.is_member(item)

    def is_member(self, item):
        "Checks whtether a item is present in a set"
        return self.__db.sismember(self.key, item)

    def remove(self, item):
        "Remove an element from set"
        self.__db.srem(self.key, item)

    def clear(self):
        """Delete set"""
        self.__db.delete(self.key)


class Counter(object):

    def __init__(self, name, value=None, namespace='counter', **redis_kwargs):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.key = '%s:%s' % (namespace, name)
        if value is not None:
            self.set(value)

    def inc(self, value=1):
        self.__db.incr(self.key, value)

    def decr(self, value=1):
        self.__db.decr(self.key, value)

    def set(self, value):
        self.__db.set(self.key, value)

    def get(self):
        value = self.__db.get(self.key)
        if value is None:
            return 0
        else:
            return int(value)


class Lock(object):
    timeout = 0

    def __init__(self, key, namespace='', expires=60, timeout=None, **redis_kwargs):
        """Distributed locking using Redis Lua scripting for CAS operations.

        Usage::

            with Lock('my_lock'):
                print "Critical section"

        :param  expires:    We consider any existing lock older than
                            ``expires`` seconds to be invalid in order to
                            detect crashed clients. This value must be higher
                            than it takes the critical section to execute.
        :param  timeout:    If another client has already obtained the lock,
                            sleep for a maximum of ``timeout`` seconds before
                            giving up. A value of 0 means we never wait.
        :param  redis:      The redis instance to use if the default global
                            redis connection is not desired.

        """
        self.key = '%s:%s' % (namespace, key)
        if timeout:
            self.timeout = timeout
        self.expires = expires
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.lock_key = None
        self._thread = None

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def acquire(self):
        """Acquire the lock

        :returns: Whether the lock was acquired or not
        :rtype: bool

        """
        if self.extend():
            return
        self.lock_key = uuid.uuid4().hex
        for _ in xrange(self.timeout + 1):
            if self.__db.setnx(self.key, self.lock_key):
                self.__db.expire(self.key, self.expires)
                self._thread = WorkerThread(self.extend, self.expires / 2)
                return
            time.sleep(1)
        raise LockTimeout("Timeout while waiting for lock")

    def is_active(self):
        return self.__db.exists(self.key)

    def extend(self):
        if self.lock_key and self.__db.get(self.key) == self.lock_key:
            self.__db.expire(self.key, self.expires)
            return True
        return False

    def release(self):
        """Release the lock

        This only releases the lock if it matches the UUID we think it
        should have, to prevent deleting someone else's lock if we
        lagged.

        """
        if self.__db.get(self.key) == self.lock_key:
            self.__db.delete(self.key)
        if self._thread and self._thread.is_alive():
            self._thread.stop()
        self.lock_key = None


class WorkerThread(threading.Thread):

    def __init__(self, target, sleep=0, *args, **kwargs):
        super(WorkerThread, self).__init__(*args, **kwargs)
        self._running = False
        self.__target = target
        self.__sleep = sleep
        self.setDaemon(True)
        self.start()

    def run(self):
        if self._running:
            return
        self._running = True
        while self._running:
            self.__target()
            time.sleep(self.__sleep)

    def stop(self):
        self._running = False
        self.join(timeout=0)


class LockTimeout(Exception):

    """Raised in the event a timeout occurs while waiting for a lock"""


class Hash(object):

    """Simple Hash with Redis Backend"""

    def __init__(self, name, namespace='hash', **redis_kwargs):
        """
        The default parameters are:
            namespace='hash', host='localhost', port=6379, db=0
        """
        self.__db = redis.Redis(
            connection_pool=poolmanager.getpool(**redis_kwargs))
        self.key = '%s:%s' % (namespace, name)

    def __len__(self):
        return self.__db.hlen(self.key)

    def __setitem__(self, idx, value):
        self.__db.hset(self.key, idx, value)

    def setnx(self, idx, value):
        return self.__db.hsetnx(self.key, idx, value) == 1

    def setifval(self, idx, oldvalue, newvalue):
        """ set ``idx`` to ``newvalue`` if current value is ``oldvalue`` """
        lua_script = """
        if redis.call('HGET', KEYS[1], ARGV[1]) == ARGV[2] then
            return redis.call('HSET', KEYS[1], ARGV[1], ARGV[3]) + 1
        else
            return nil
            end
        """
        return self.__db.eval(lua_script, 1, self.key, idx, oldvalue, newvalue)

    def __getitem__(self, key):
        value = self.__db.hget(self.key, key)
        return int(value) if isinstance(value, basestring) and value.isdigit() else value

    def inc(self, key, value=1):
        return self.__db.hincrby(self.key, key, value)

    def __iter__(self):
        return self.__db.hscan_iter(self.key)

    def scan(self, match=None):
        return self.__db.hscan_iter(self.key, match=match)

    def clear(self):
        return self.__db.delete(self.key)

    def keys(self):
        return self.__db.hkeys(self.key)

    def values(self):
        intify = lambda value: int(value) if isinstance(value, basestring) and value.isdigit() else value
        return map(intify, self.__db.hvals(self.key))




REDIS_URL = 'localhost'
REDIS_PORT = 6379
REDIS_DB = 1

TIME_ZONE = 'UTC'
RESUME = False
MAX_RETRY = 3
UNIQUE_CHECK = True
THREADS = 5

AUTOTHROTTLE = False
TIMEOUT = 5
MIN_DELAY = 0.5
MAX_DELAY = 60
REQUEST_PROCESSOR = 'dragline.http:RequestProcessor'

DEFAULT_REQUEST_ARGS = {
    'allow_redirects': True,
    'auth': None,
    'cert': None,
    'cookies': None,
    'data': None,
    'files': None,
    'headers': None,
    'json': None,
    'method': None,
    'params': None,
    'proxies': None,
    'stream': False,
    'timeout': None,
    'verify': False
}

SELENIUM_ARGS = {
    'proxy': None,
    'WINDOW_SIZE_WIDTH': 1024,
    'WINDOW_SIZE_HEIGHT': 768,
    'DISABLE_CSS': False,
    'DISABLE_IMAGE': False,
    'DISABLE_FLASH': False
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'standard'
        }
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': False
        },
        'dragline': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': False
        }
    }
}



from pprint import pformat
from collections import MutableMapping
from abc import ABCMeta
import six
from datetime import datetime
import json


class BaseItem(object):
    """Base class for all scraped items."""
    pass


class Field(object):
    primary = False
    """Container of field metadata"""
    def __init__(self, **args):
        for k, v in args.items():
            setattr(self, k, v)

    def validate(self, key, value):
        return value

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class TextField(Field):
    type = 'text'

    def validate(self, key, value):
        if not isinstance(value, basestring):
            raise TypeError('got value of {} expected {} for {}'.format(type(value), 'basestring', key))
        return value


class IntField(Field):
    type = 'int'

    def validate(self, key, value):
        try:
            value = int(value)
        except:
            raise TypeError('got value of {} expected {} for {}'.format(type(value), self.type, key))
        return value


class DecimalField(Field):
    type = 'double'

    def validate(self, key, value):
        try:
            value = float(value)
        except:
            raise TypeError('got value of {} expected {} for {}'.format(type(value), 'float', key))
        return value


class JSONField(Field):
    type = 'text'

    def validate(self, key, value):
        return value


class DatetimeField(Field):
    type = 'timestamp'

    def validate(self, key, value):
        if not isinstance(value, datetime):
            raise TypeError('got value of {} expected {} for {}'.format(type(value), 'datetime', key))
        return value


class ItemMeta(ABCMeta):

    def __new__(mcs, class_name, bases, attrs):
        fields = {}
        new_attrs = {}
        for n, v in six.iteritems(attrs):
            if isinstance(v, Field):
                fields[n] = v
            else:
                new_attrs[n] = v

        cls = super(ItemMeta, mcs).__new__(mcs, class_name, bases, new_attrs)
        cls.fields = cls.fields.copy()
        cls.fields.update(fields)
        return cls


class DictItem(MutableMapping, BaseItem):

    fields = {}

    def __init__(self, *args, **kwargs):
        self._values = {}
        if args or kwargs:  # avoid creating dict for most common case
            for k, v in six.iteritems(dict(*args, **kwargs)):
                self[k] = v

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        if key in self.fields:
            if value is None:
                self._values[key] = None
            else:
                self._values[key] = self.fields[key].validate(key, value)
        else:
            raise KeyError("%s does not support field: %s" %
                (self.__class__.__name__, key))

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name in self.fields:
            raise AttributeError("Use item[%r] to get field value" % name)
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            raise AttributeError("Use item[%r] = %r to set field value" % (name, value))
        super(DictItem, self).__setattr__(name, value)

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    __hash__ = BaseItem.__hash__

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return pformat(dict(self))

    def copy(self):
        return self.__class__(self)


@six.add_metaclass(ItemMeta)
class Item(DictItem):
    def save(self, *args, **kwargs):
        print dict(self)




from __future__ import absolute_import
from requests.compat import urlencode, urldefrag
import socket
from hashlib import sha1
import requests
from requests.structures import CaseInsensitiveDict
import json
import types
import sys
from dragline import runtime

try:
    from requests.packages import urllib3
    urllib3.disable_warnings()
except:
    pass
socket.setdefaulttimeout(300)


class RequestError(Exception):
    pass


class Request(object):

    """
    :param url: URL to send.
    :param method: HTTP method to use.
    :param callback: name of the function to call after url is downloaded.
    :param meta:  A dict that contains arbitrary metadata for this request.
    :param headers: dictionary of headers to send.
    :param files: dictionary of {filename: fileobject} files to multipart upload.
    :param data: the body to attach to the request. If a dictionary is provided, form-encoding will take place.
    :param json: json for the body to attach to the request (if data is not specified).
    :param params: dictionary of URL parameters to append to the URL.
    :param auth: Auth handler or (user, pass) tuple.
    :param cookies: dictionary or CookieJar of cookies to attach to this request.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a tuple.
    :type timeout: float or tuple
    :param allow_redirects: (optional) Set to True by default.
    :type allow_redirects: bool
    :param proxies: (optional) Dictionary mapping protocol to the URL of
        the proxy.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) if ``True``, the SSL cert will be verified.
        A CA_BUNDLE path can also be provided.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair.
    """
    callback = None
    meta = None
    dont_filter = False
    retry = 0

    def __init__(self, url, method=None, callback=None, meta=None, headers=None, files=None, data=None, json=None,
                 params=None, auth=None, cookies=None, timeout=None, allow_redirects=None, proxies=None, stream=None,
                 verify=None, cert=None, dont_filter=None):
        args = dict(url=url, method=method, headers=headers, files=files, data=data, json=json, params=params,
                    auth=auth, cookies=cookies, timeout=timeout, allow_redirects=allow_redirects, proxies=proxies,
                    stream=stream, verify=verify, cert=cert)
        self.request_args = {k: v for k, v in args.items() if v is not None}
        if callback is not None:
            self.callback = callback
        if meta is not None:
            self.meta = meta
        if dont_filter:
            self.dont_filter = True

    def __getstate__(self):
        d = self.__dict__.copy()
        if isinstance(self.callback, types.MethodType) and hasattr(self.callback, 'im_self'):
            d['callback'] = self.callback.__name__
            if not runtime.spider == self.callback.im_self:
                runtime.spider = self.callback.im_self
        return d

    def __setstate__(self, d):
        if 'callback' in d and isinstance(d['callback'], str):
            d['callback'] = getattr(runtime.spider, d['callback'])
        self.__dict__ = d

    def __repr__(self):
        return "<%s>" % self.get_unique_id()

    def __str__(self):
        return self.get_unique_id()

    def __usha1(self, x):
        """sha1 with unicode support"""
        if isinstance(x, unicode):
            return sha1(x.encode('utf-8')).hexdigest()
        else:
            return sha1(x).hexdigest()

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError as e:
            return self.request_args[name]
        raise e

    def get_args(self, keys=None):
        args = runtime.settings.DEFAULT_REQUEST_ARGS.copy()
        args.update(self.request_args)
        if args['method'] is None:
            if any((args['data'], args['json'])):
                args['method'] = "POST"
            else:
                args['method'] = "GET"
        # if 'timeout' not in args:
        #    args['timeout'] = max(self.settings.DELAY, self.settings.TIMEOUT)
        for i in args:
            if callable(args[i]):
                args[i] = args[i]()
        if keys:
            return {k: args[k] for k in keys}
        return args

    def send(self):
        """
        This function sends HTTP requests.

        :returns: response
        :rtype: :class:`dragline.http.Response`
        :raises: :exc:`dragline.http.RequestError`: when failed to fetch contents

        >>> req = Request("http://www.example.org")
        >>> response = req.send()
        >>> print response.headers['status']
        200

        """
        try:
            response = runtime.request_processor.get_response(**self.get_args())
            response.meta = self.meta
            return response
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except:
            exc_class, exc, tb = sys.exc_info()
            new_exc = RequestError("Failed to send Request on %s: %s" % (self, exc))
            raise new_exc.__class__, new_exc, tb

    def get_unique_id(self, hashing=False):
        keys = ["method", "url", "data", "params", "json"]
        args = self.get_args(keys)
        args["url"] = urldefrag(args["url"])[0]
        request = ":".join(args[k] for k in keys[:2])
        if args["params"]:
            request += "?" + self._encode_params(args["params"])
        body = None
        if args["json"]:
            body = json.dumps(args["json"])
        elif args["data"]:
            body = self._encode_params(args["data"])
        if body:
            request += ":" + body
        if hashing:
            return self.__usha1(request)
        else:
            return request

    @staticmethod
    def _encode_params(data):
        """Encode parameters in a piece of data.

        Will successfully encode parameters when passed as a dict or a list of
        2-tuples. Order is retained if data is a list of 2-tuples but arbitrary
        if parameters are supplied as a dict.
        """
        if data is None:
            return ''
        elif isinstance(data, basestring):
            return data
        elif hasattr(data, 'read'):
            return data
        elif hasattr(data, '__iter__'):
            result = []
            for k, vs in sorted(CaseInsensitiveDict(data).lower_items()):
                if isinstance(vs, basestring) or not hasattr(vs, '__iter__'):
                    vs = [vs]
                for v in vs:
                    if v is not None:
                        result.append(
                            (k.encode('utf-8') if isinstance(k, str) else k,
                             v.encode('utf-8') if isinstance(v, str) else v))
            return urlencode(result, doseq=True)
        else:
            return data


class RequestProcessor(object):
    def get_response(self, **kwargs):
        session = requests.Session()
        response = session.request(**kwargs)

        if response.encoding == 'ISO-8859-1' and 'ISO-8859-1' not in response.headers.get('Content-Type', ''):
                response.encoding = response.apparent_encoding
        response.cookies = session.cookies
        return Response(response)

    def put_response(self, response):
        del response

    def clear(self):
        pass


class Response(requests.Response):
    """The :class:`Response <Response>` object, which contains a
    server's response to an HTTP request.
    """

    __attrs__ = [
        '_content',
        'status_code',
        'headers',
        'url',
        'history',
        'encoding',
        'reason',
        'cookies',
        'elapsed',
        'request',
        'meta'
    ]

    meta = None

    def __init__(self, response=None):
        if isinstance(response, requests.Response):
            self.__dict__ = response.__dict__

    def __len__(self):
        if 'content-length' in self.headers:
            return int(self.headers['content-length'])
        return self.raw.tell()

    @property
    def body(self):
        return self.content

    @property
    def status(self):
        return self.status_code



from lxml import html, etree
from funcs import xpathstrip, xpathtostring, extract_html, extract_text
from urlparse import urlsplit


ns = etree.FunctionNamespace(None)
ns['strip'] = xpathstrip
ns['str'] = xpathtostring


class HTMLElement(html.HtmlMixin):
    """
    HtmlElement object is returned by the HtmlParser function:

    >>> response = Request('http://www.example.org/').send()
    >>> parser = HtmlParser(response)
    """

    def extract_urls(self, xpath=None, domains=None):
        """
        Returns a list of all the links with given domains
        """
        if xpath and not xpath.endswith('/'):
            xpath += '/'
        elif xpath is None:
            xpath = ""
        for url in map(urlsplit, set(self.xpath(xpath + "descendant-or-self::a/@href"))):
            if url.scheme in ['http', 'https'] and url.hostname:
                if domains and url.hostname not in domains:
                    continue
                yield url.geturl()

    def extract_text(self):
        """
        Returns the text content of the element,
        including the text content of its children, with no markup.

        >>> list(parser.extract_urls())
        ['http://www.iana.org/domains/example']
        """

        return extract_text(self)

    def html_content(self):
        return extract_html(self)

    def x(self, xpaths, default=None):
        if isinstance(xpaths, basestring):
            xpaths = [xpaths]
        for xpath in xpaths:
            result = self.find('*' + xpath)
            if result is not None:
                return result
        return default


class HtmlComment(etree.CommentBase, HTMLElement):
    pass


class HtmlElement(etree.ElementBase, HTMLElement):
    pass


class HtmlProcessingInstruction(etree.PIBase, HTMLElement):
    pass


class HtmlEntity(etree.EntityBase, HTMLElement):
    pass


class HtmlElementClassLookup(html.HtmlElementClassLookup):

    def lookup(self, node_type, document, namespace, name):
        if node_type == 'element':
            return self._element_classes.get(name.lower(), HtmlElement)
        elif node_type == 'comment':
            return HtmlComment
        elif node_type == 'PI':
            return HtmlProcessingInstruction
        elif node_type == 'entity':
            return HtmlEntity
        # Otherwise normal lookup
        return None


class HTMLParser(etree.HTMLParser):

    """An HTML parser that is configured to return lxml.html Element
    objects.
    """

    def __init__(self, **kwargs):
        super(HTMLParser, self).__init__(**kwargs)
        self.set_element_class_lookup(HtmlElementClassLookup())


def HtmlParser(response, absolute_links=True):
    """
    :param response:
    :type response: :class:`dragline.http.Response`

    This method takes response object as its argument and returns
    the lxml etree object.

    HtmlParser function returns a lxml object of type HtmlElement which got few potential methods.
    All the details of lxml object are discussed in section `lxml.html.HtmlElement`.
    """
    encoding = response.encoding if hasattr(response, 'encoding') else 'utf-8'
    parser = HTMLParser(recover=True, encoding=encoding)
    if isinstance(response, basestring):
        element = html.fromstring(response, None, parser)
    else:
        element = html.fromstring(response.content, response.url, parser)
        if absolute_links:
            element.make_links_absolute()
    return element



from glob import glob
from setuptools import setup, find_packages


setup(
    name='Dragline',
    version=open('dragline/VERSION').read().strip(),
    description='Distributed Python web crawling framework',
    author='Ashwin Rajeev, Shimil Rahman',
    author_email='ashwin@inzyte.com, shimil@inzyte.com',
    url='http://www.inzyte.com',
    packages=find_packages(exclude=['docs', 'tests*']),
    scripts=glob("scripts/*"),
    data_files=[('dragline/templates', glob("dragline/templates/*"))],
    include_package_data=True,
    install_requires=[i.strip() for i in open('requirements.txt')],
    test_suite='tests',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP']
)
